# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Perform "hardcover" imposition."""

from pdfimpose.schema import hardcover

from .common import (
    parse_bind,
    parse_format,
    parse_group,
    parse_last,
    parse_length,
    parse_marks,
    parse_repeat,
    parse_resize,
)


def impose(infile, outfile, arguments):
    """Perform imposition.

    :param dict arguments: Arguments got from the POST html method.
    """
    parameters = {}

    # Get paper size
    key, value = parse_format(arguments, "form-hardcover")
    parameters[key] = value

    # Argument "last"
    parameters["last"] = parse_last(arguments, "form-hardcover")

    # Arguments "imargin" and "omargin"
    parameters["imargin"] = parse_length(arguments, "form-hardcover", "imargin")
    parameters["omargin"] = parse_length(arguments, "form-hardcover", "omargin")

    # Argument "mark"
    parameters["mark"] = parse_marks(arguments, "form-hardcover", ("bind", "crop"))

    # Argument "bind"
    parameters["bind"] = parse_bind(arguments, "form-hardcover")

    # Argument "group"
    parameters["group"] = parse_group(arguments, "form-hardcover")

    # Argument "resize"
    infile = parse_resize(infile, arguments, "form-hardcover")

    hardcover.impose(
        files=[infile for _ in range(parse_repeat(arguments, "form-hardcover"))],
        output=outfile,
        **parameters,
    )



# Copyright 2024-2025 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Web interface to pdfimpose https://framagit.org/spalax/pdfimpose"""

import atexit
import collections
import contextlib
import datetime
import functools
import operator
import os
import pathlib
import secrets
import shutil
import tempfile
import time
import typing

import papersize
import pdfimpose
from flask import (
    Flask,
    flash,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    send_from_directory,
    session,
    url_for,
)
from flask_babel import Babel
from flask_babel import lazy_gettext as _
from flask_basicauth import BasicAuth
from werkzeug.exceptions import HTTPException
from werkzeug.utils import secure_filename

from . import clean, database, layout, limiter, settings

VERSION = "1.3.0"


def allowed_file(filename: str):
    """Return `True` if the filename is allowed."""
    return filename.endswith(".pdf")


def papersizetranslations():
    """Return the location of translation of package `papersize`."""
    # Thanks to:
    # https://importlib-resources.readthedocs.io/en/latest/migration.html#pkg-resources-resource-filename
    file_manager = contextlib.ExitStack()
    atexit.register(file_manager.close)
    return file_manager.enter_context(papersize.translation_directory())


@functools.cache
def get_languages(babel: Babel) -> collections.OrderedDict[str, str]:
    """Return an ordered dictionary of available languages, sorted by language code.

    - Keys are the language codes (e.g. "en" for English).
    - Values are the display names (e.g. "English").
    """
    return collections.OrderedDict(
        (locale.language, locale.language_name)
        for locale in sorted(
            set(babel.list_translations()), key=operator.attrgetter("language")
        )
    )


# pylint: disable=too-many-statements
def create_app():
    """Create and configure the application"""

    app = Flask(__name__, instance_relative_config=True)
    app.jinja_env.add_extension("jinja2.ext.loopcontrols")

    # Load options
    settings.set_configuration(app)

    app.config["SESSION_COOKIE_SAMESITE"] = "Strict"

    # Configure Babel
    app.config["BABEL_TRANSLATION_DIRECTORIES"] = (
        f"translations;{papersizetranslations()}"
    )
    app.config["BABEL_DOMAIN"] = "messages;papersize"

    def get_locale():
        """Return the chosen locale.

        - If a locale has been passed as argument '?language=en', and a translation is available:
          use it, and store the choice in cookies.
        - Otherwise, is a valid language has been stored in cookies, use it.
        - Otherwise, guess it.
        """
        # Has locale been passed as an argument '?language=en'?
        if "language" in request.args:
            if request.args["language"] in get_languages(babel):
                return request.args["language"]

        # Has locale been stored in cookies?
        if request.cookies.get("language") in get_languages(babel):
            return request.cookies.get("language")

        # Guess locale
        return request.accept_languages.best_match(
            locale.language for locale in babel.list_translations()
        )

    babel = Babel(app, locale_selector=get_locale)

    # Configure database
    database.init(app)

    # Configure limiter
    limiter.init(app)

    # Create temporary folder
    app.config["TEMP_FOLDER"].mkdir(parents=True, exist_ok=True)

    # Clean files from older runs, and create cleaner (that removes PDFs)
    if app.config["RUN_CLEANER_THREAD"]:
        toremove = clean.launch_thread_cleaner(app)

    # Add stuff to template context

    @app.context_processor
    def context_processor():
        return {
            "UNITS": {
                key: _(value).split("(")[0].strip()
                for key, value in papersize.UNITS_HELP.items()
                if key in ("pt", "mm", "cm", "in")
            },
            "SIZES": {key: _(value) for key, value in papersize.SIZES_HELP.items()},
            "LAYOUTS": (
                # Layouts are sorted from a beginner point-of-view:
                # the simplest, most usual, to the weirdest.
                "saddle",
                "cards",
                "pdfautonup",
                "hardcover",
                "onepagezine",
                "wire",
                "copycutfold",
                "cutstackfold",
            ),
            "languages": get_languages(app.extensions["babel"].instance),
            "lang": app.extensions["babel"].locale_selector(),
            "stat": database.stat(),
            "max_size": database.prettyprint_size(
                app.config["MAX_CONTENT_LENGTH"], round_func=round
            ),
        }

    # Quick and dirty statistics page
    basic_auth = BasicAuth(app)

    # Redirect while keeping query string
    def smart_redirect(location, *args, **kwargs):
        if request.query_string:
            return redirect(
                f"{location}?{request.query_string.decode()}", *args, **kwargs
            )
        return redirect(location, *args, **kwargs)

    # Define routes

    @app.route("/stats")
    @basic_auth.required
    def stats():
        return render_template(
            "stats.html",
            dataday=database.get_history("day", dateformat='"%Y-%m-%d"'),
            dataweek=database.get_history("week", dateformat='"%Y-%m-%d"'),
            datamonth=database.get_history("month", dateformat='"%Y-%m"'),
            datayear=database.get_history("year", dateformat='"%Y"'),
        )

    @app.route("/", methods=["GET", "POST"])
    def root():
        if request.method == "POST":
            # Catch argument errors
            try:
                if (
                    app.config["MAX_FILE_REPEAT"]
                    != 0  # MAX_FILE_REPEAT == 0 means no limit
                    and int(
                        request.form.to_dict()[
                            f"""form-{request.form["layout"]}-repeat-value"""
                        ]
                    )
                    > app.config["MAX_FILE_REPEAT"]
                ):
                    flash(
                        _("Error: The maximum number of repetitions is %s.")
                        % app.config["MAX_FILE_REPEAT"],
                        category="impose",
                    )
                    return smart_redirect("/")
            except (KeyError, ValueError):
                pass

            # Get raw list of (possibly bad) file names
            rawfiles = request.files.getlist("file")

            # Remove bad files, save good files
            tempdir = pathlib.Path(
                tempfile.mkdtemp(
                    prefix=datetime.datetime.now().strftime("%Y%m%d-%H%M%S-"),
                    dir=app.config["TEMP_FOLDER"],
                )
            )
            uploaddir = tempdir / "upload"
            uploaddir.mkdir(parents=True, exist_ok=True)

            sourcefiles = []
            totalsize = 0
            for file in rawfiles:
                if not file.filename:
                    flash(_("Error: No file."), "files")
                    continue
                if not allowed_file(file.filename):
                    flash(_("Error: Invalid filename."), "files")
                    continue
                sourcefiles.append(pathlib.Path(secure_filename(file.filename)))
                file.save(uploaddir / sourcefiles[-1])
                totalsize += (uploaddir / sourcefiles[-1]).stat().st_size

            # No source files…
            if not sourcefiles:
                flash(
                    _("No valid PDF files found. Try uploading a valid PDF file."),
                    "files",
                )
                return smart_redirect("/")

            # Create download directory
            downloaddir = tempdir / "download"
            downloaddir.mkdir(parents=True, exist_ok=True)
            destfile = f"{sourcefiles[0].stem}-impose{sourcefiles[0].suffix}"

            # Impose file
            try:
                layout.impose(
                    request.form["layout"],
                    infile=uploaddir / sourcefiles[0],
                    outfile=downloaddir / destfile,
                    arguments=request.form.to_dict(),
                )
            except pdfimpose.UserError as error:
                flash(  #  pylint: disable=consider-using-f-string
                    _("Error while imposing files: %s") % error,
                    category="impose",
                )
                return smart_redirect("/")

            # Imposition succeeded. Log it in the database
            database.add(totalsize)

            # Remove source files, and mark produced file for removal
            shutil.rmtree(uploaddir, ignore_errors=True)
            if app.config["RUN_CLEANER_THREAD"]:
                toremove.put(
                    (datetime.datetime.now() + app.config["PDF_LIFETIME"], tempdir)
                )

            # Everything went right!
            return send_from_directory(
                downloaddir,
                destfile,
                as_attachment=True,
            )
        # Method GET (or anything but POST)
        if "noscript" in request.args:
            template = "noscript.html"
        else:
            template = "index.html"
        response = make_response(render_template(template))

        if request.args.get("language", default=None) in get_languages(babel):
            # Store preferred language in cookies
            response.set_cookie("language", request.args["language"])

        return response

    @app.errorhandler(413)
    def error413(error):
        #  pylint: disable=unused-argument
        flash(
            _(
                "Error: File is too big: maximum size is %s.",
            )
            % database.prettyprint_size(
                app.config["MAX_CONTENT_LENGTH"], round_func=round
            ),
            "files",
        )
        return render_template("index.html"), 413

    return app



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Delete temporary files older than one day."""

import datetime
import pathlib
import queue
import re
import shutil
import threading
import time

RE_DATE = re.compile("^[0-9]{8}-[0-9]{6}-")


def clean(download: pathlib.Path, lifetime: datetime.timedelta):
    """Remove outdated files from the download folder.

    :param pathlib.Path download: Directory to clean.
    :param datetime.timedelta lifetime: Name of files older than this are removed.
    :return: Name of folders that are too young to be removed, together with their removal datetime.
    :rtype: Tuple[pathlib.Path, datetime.datetime]
    """
    young = set()
    for folder in download.iterdir():
        if folder.is_dir() and (match := RE_DATE.match(str(folder.name))):
            deathtime = (
                datetime.datetime.strptime(match.group(), "%Y%m%d-%H%M%S-") + lifetime
            )
            if deathtime < datetime.datetime.now():
                shutil.rmtree(download / folder.name, ignore_errors=True)
            else:
                young.add((download / folder.name, deathtime))
    return young


def _thread_cleaner(toremove: queue.Queue):
    """Read names from the queue, and remove the corresponding files."""
    while True:
        date, path = toremove.get()
        time.sleep(max(0, (date - datetime.datetime.now()).total_seconds()))
        shutil.rmtree(path, ignore_errors=True)


def launch_thread_cleaner(app) -> queue.Queue:
    """Launch the cleaner thread, and return a queue of files to clean."""
    toremove = queue.Queue()
    for folder, deathtime in clean(
        app.config["TEMP_FOLDER"], app.config["PDF_LIFETIME"]
    ):
        toremove.put((deathtime, folder))
    threading.Thread(
        target=_thread_cleaner,
        args=(toremove,),
        daemon=True,
    ).start()
    return toremove



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Default settings for the application."""

import datetime
import secrets

# Set the temporary folder (in which user files and imposed files will be temporary stored).
# If `None`, a default one will be used.
# Example: TEMP_FOLDER = "/path/to/temporary/folder"
TEMP_FOLDER = None

# Maximum size (in bytes) of files that can be processed by the application.
MAX_CONTENT_LENGTH = 50_000_000

# Maximum number of times files are repeated.
# Set to 0 to allow arbitrary big numbers. Do not use 0 in production!
# Users could generate arbitrarily big PDF files, that will clog your server
# CPU, and fill your server memory.
MAX_FILE_REPEAT = 20

# Secret key used for securely signing default cookies.
# By default, a new one will be chosen at random each time the application restarts,
# which means users might experince bugs if you restart the application while they are browsing.
# Example: SECRET_KEY = 'da5ad311558575db9fbd534fd1fc34b1'
SECRET_KEY = secrets.token_hex(16)

# Set how long uploaded PDF files and imposed files should live before being removed,
# either by the cleaner thread (see below) or by manually calling ``pdfimposeweb.clean``.
# Example: PDF_LIFETIME = datetime.timedelta(hours=1)
PDF_LIFETIME = datetime.timedelta(hours=2)

# If ``True``, at the application startup, a thread is started to automatically
# remove outdated PDF files (depending on the `PDF_LIFETIME` option).
# If ``False``, one has to manually run ``pdfimposeweb.clean`` to remove
# outdated PDF files (in a crontab for instance).
RUN_CLEANER_THREAD = True

# Database URI. See the flask-sqlalchemy documentation to see how to connect to your database.
# https://flask-sqlalchemy.palletsprojects.com/en/3.0.x/config/#connection-url-format
# By default, an in-memory sqlite database is used
# (which is reset everytime the application restarts).
SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"

# Username and password to view the statistics page http://YOUR-DOMAIN-NAME/stats
# This is a quick-and-dirty page, with a quick-and-dirty authentication method.
# The only reason this page is protected is because it is badly coded, and
# might stress the database.
# By default, random username and password are chosen (basically forbidding
# anybody to see the page).
# This is experimental, and might be broken in later versions without notice.
BASIC_AUTH_USERNAME = secrets.token_hex(16)
BASIC_AUTH_PASSWORD = secrets.token_hex(16)

# Options for flask-limiter, which adds rate limiting to pdfimposeweb.
# Those are only some of the option that are recognisez by `flask_limiter`.
# See the full list of available options here:
# https://flask-limiter.readthedocs.io/en/stable/configuration.html#using-flask-config
# By default, data is stored in memory, but according to the `flask-limiter` documentation,
# this is designed for development and testing purpose, and should not be used in production.
RATELIMIT_DEFAULT = "200 per day, 50 per hour"
RATELIMIT_STORAGE_URI = "memory://"
RATELIMIT_STORAGE_OPTIONS = {}



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Common functions related to layout (mainly about parsing arguments)."""

from pdfimpose import DEFAULT_PAPER_SIZE

from .. import pdfresize


def parse_format(arguments, prefix, *, default=DEFAULT_PAPER_SIZE):
    """Parse arguments about paper size."""
    try:
        if arguments[f"{prefix}-format"] == "standard":
            return "size", arguments[f"{prefix}-format-standard"]
        if arguments[f"{prefix}-format"] == "custom":
            return (
                "size",
                "{}{}x{}{}".format(  #  pylint: disable=consider-using-f-string
                    arguments[f"{prefix}-format-custom-width-value"],
                    arguments[f"{prefix}-format-custom-width-unit"],
                    arguments[f"{prefix}-format-custom-height-value"],
                    arguments[f"{prefix}-format-custom-height-unit"],
                ),
            )
        if arguments[f"{prefix}-format"] == "signature":
            return "signature", (
                int(arguments[f"{prefix}-format-signature-width"]),
                int(arguments[f"{prefix}-format-signature-height"]),
            )
    except (KeyError, ValueError):
        return default
    return default


def parse_length(arguments, prefix, key, *, default=0):
    """Parse argument about length (value + unit)."""
    try:
        return "{}{}".format(  #  pylint: disable=consider-using-f-string
            arguments[f"{prefix}-{key}-value"],
            arguments[f"{prefix}-{key}-unit"],
        )
    except KeyError:
        return default


def parse_marks(arguments, prefix, marks, *, default=None):
    """Parse argument about crop and bind marks."""
    try:
        enabled = []
        for mark in marks:
            if arguments[f"{prefix}-marks-{mark}"] == "on":
                enabled.append(mark)
        return enabled
    except KeyError:
        return default


def parse_last(arguments, prefix, *, default=0):
    """Parse arguments about last pages."""
    try:
        return int(arguments[f"{prefix}-last"])
    except (KeyError, ValueError):
        return default


def parse_bind(arguments, prefix, *, default="left"):
    """Parse arguments about binding edge."""
    try:
        return arguments[f"{prefix}-bind"]
    except KeyError:
        return default


def parse_group(arguments, prefix, *, default=1):
    """Parse argument about grouping sheets of paper."""
    try:
        if arguments[f"{prefix}-group"] == "some":
            return int(arguments[f"{prefix}-group-value"])
        if arguments[f"{prefix}-group"] == "no":
            return 1
        if arguments[f"{prefix}-group"] == "all":
            return 0
    except (KeyError, ValueError):
        return default
    return default


def parse_repeat(arguments, prefix):
    """Parse argument about input repetition."""
    try:
        return int(arguments[f"{prefix}-repeat"])
    except (KeyError, ValueError):
        return 1


def parse_resize(infile, arguments, prefix):
    """Parse argument about resizing input, and perform resize.

    Return the input file name (original name if no resizing is done, resized
    file if file has been resized).
    """
    newfile = infile.with_stem(f"{infile.stem}-resized")
    try:
        if arguments[f"{prefix}-resize"] == "dont":
            return infile
        if arguments[f"{prefix}-resize"] == "standard":
            pdfresize.resize_size(
                infile, newfile, arguments[f"{prefix}-resize-standard"]
            )
            return newfile
        if arguments[f"{prefix}-resize"] == "custom":
            pdfresize.resize_size(
                infile,
                newfile,
                "{}{}x{}{}".format(  #  pylint: disable=consider-using-f-string
                    arguments[f"{prefix}-resize-custom-width-value"],
                    arguments[f"{prefix}-resize-custom-width-unit"],
                    arguments[f"{prefix}-resize-custom-height-value"],
                    arguments[f"{prefix}-resize-custom-height-unit"],
                ),
            )
            return newfile
        if arguments[f"{prefix}-resize"] == "scale":
            pdfresize.resize_scale(
                infile, newfile, float(arguments[f"{prefix}-resize-scale"])
            )
            return newfile
        return infile
    except (KeyError, ValueError):
        return infile



# Copyright 2024-2025 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Define and process database."""

# pylint: disable=too-few-public-methods

import datetime
import functools
import logging
import math
import typing

from flask_babel import lazy_pgettext as pgettext
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import exc, func
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column


class Base(MappedAsDataclass, DeclarativeBase):
    """Base SQLAlchemy class."""


db = SQLAlchemy(model_class=Base)


class DailyStat(db.Model):
    """Statistics about one single day.

    - How many files were processed that day?
    - What were their total size (in bytes)?
    """

    date: Mapped[datetime.date] = mapped_column(
        primary_key=True, default_factory=datetime.date.today
    )
    files: Mapped[int] = mapped_column(default=0)
    size: Mapped[int] = mapped_column(default=0)


def init(app):
    """Initialize the database.

    If the tables do not exist, they are created.
    """
    if "SQLALCHEMY_ENGINE_OPTIONS" not in app.config:
        app.config["SQLALCHEMY_ENGINE_OPTIONS"] = {}
    app.config["SQLALCHEMY_ENGINE_OPTIONS"].update({"pool_recycle": 280})
    db.init_app(app)

    with app.app_context():
        db.create_all()
        if not db.session.query(DailyStat).first():
            # Database is empty.
            # Create an empty row: it will be used to know the date the application was started on.
            db.session.add(DailyStat())
            db.session.commit()


def add(size: int):
    """Record the information that a new file has been processed.

    :param int size: The file size, in bytes.

    The database is changed to record that:
    *today*, *one* more file of *size* Mb has been processed.
    """
    # Avoiding the weird condition where today() changes *while running this function*.
    today = datetime.date.today()

    if db.session.get(DailyStat, today) is None:
        try:
            db.session.add(DailyStat())
            db.session.commit()
        except exc.IntegrityError:
            # Today statistics has already been created by another process.
            db.session.rollback()

    todaystat = db.session.get(DailyStat, today)
    todaystat.files = DailyStat.files + 1
    todaystat.size = DailyStat.size + size
    db.session.commit()


_UNITS = [
    pgettext("unit", "b"),
    pgettext("unit", "kb"),
    pgettext("unit", "Mb"),
    pgettext("unit", "Gb"),
    pgettext("unit", "Tb"),
    pgettext("unit", "Pb"),
]


def _between(low, number, hight):
    return min(max(number, low), hight)


def prettyprint_size(size: int, *, round_func: typing.Callable = None) -> str:
    """Return a human readable size.

    :param int size: Size, in bytes.
    :return: A human readable size.

    >>> prettyprint_size(12345)
    l'12.345 kb'
    >>> prettyprint_size(51_234_567, round_func=round)
    l'51 Mb'
    """
    if size == 0:
        power3 = 0
    else:
        power3 = _between(0, math.floor(math.log(size, 1000)), len(_UNITS) - 1)
    if round_func is None:
        round_func = lambda x: x  #  pylint: disable=unnecessary-lambda-assignment

    return pgettext(
        "unit",
        "%(size)s %(unit)s",
        size=round_func(size / 10 ** (3 * power3)),
        unit=_UNITS[power3],
    )


def ttl_cache(delta: datetime.timedelta):
    """A function cache that calls the actual function once every `delta` time.

    Only works with functions that take no arguments.
    """

    def wrapper(function: typing.Callable):
        # The idea is to use :func:`functools.lru_cache` with a dummy argument
        # that changes each day. Idea borrowed from Peng Qian
        # https://www.dataleadsfuture.com/implement-a-cache-decorator-with-time-to-live-feature-in-python/

        def count() -> int:
            """Increment value each day"""
            start = datetime.datetime.now()
            while True:
                yield math.floor((datetime.datetime.now() - start) / delta)

        counter = count()

        @functools.lru_cache(maxsize=1)
        def lru_cached_func(dummy):  #  pylint: disable=unused-argument
            """Dummy function to make :func:`functools.lru_cache` handle cache.

            The dummy argument is changed every time we want to run the
            function instead of returning the cached data.
            """
            return function()

        def wrapped():
            return lru_cached_func(next(counter))

        return wrapped

    return wrapper


@ttl_cache(datetime.timedelta(days=1))
def stat() -> dict:
    """Return some statistics about this application since it was started.

    Data is:
    - number of files processed;
    - year it was started on;
    - total size of processed files.

    Might return an empty dictionary if an exception is raised while querying the database.
    """
    try:
        files, size = db.session.execute(
            db.select(func.sum(DailyStat.files), func.sum(DailyStat.size))
        ).first()
        year = (
            db.session.execute(db.select(DailyStat.date).order_by("date"))
            .first()[0]
            .year
        )
    except exc.SQLAlchemyError as error:
        logging.warning("Error while querying statistics from the database: %s", error)
        return {}
    return {
        "files": files,
        "year": year,
        "size": prettyprint_size(size, round_func=math.floor),
    }


# Dialect-agnostic function to format date
def strftime(dialect, date, dateformat):
    """Format date according to dialect."""
    if dialect == "sqlite":
        return func.strftime(dateformat, date)
    if dialect == "mysql":
        return func.date_format(date, dateformat)
    if dialect == "postgresql":
        return func.to_char(date, dateformat)
    raise NotImplementedError(
        f"Dialect {dialect} is not supported. Please submit an issue or a patch."
    )


def _iter_months(start, end):
    """Iter first days of month from start to end.

    The first item is the first day of the month of `start`.
    The last item is the first day of the month of `end`.

    >>> list(_iter_months(datetime.date(2024, 11, 10), datetime.date(2025, 1, 7)))
    [datetime.date(2024, 11, 1), datetime.date(2024, 12, 1), datetime.date(2025, 1, 1)]
    """
    month = start.replace(day=1)
    while month <= end:
        yield month
        month = (month + datetime.timedelta(days=31)).replace(day=1)


def _iter_weeks(start, end):
    """Iter first days of week from start to end.

    The first item is the first day of the week of `start`.
    The last item is the first day of the week of `end`.

    >>> list(_iter_weeks(datetime.date(2023, 12, 20), datetime.date(2024, 1, 7)))
    [datetime.date(2023, 12, 18), datetime.date(2023, 12, 25), datetime.date(2024, 1, 1)]
    """
    week = start - datetime.timedelta(days=start.weekday() % 7)
    while week <= end:
        yield week
        week += datetime.timedelta(days=7)


def _iter_days(start, end):
    """Iter days from start to end.

    >>> list(_iter_days(datetime.date(2023, 12, 31), datetime.date(2024, 1, 2)))
    [datetime.date(2023, 12, 31), datetime.date(2024, 1, 1), datetime.date(2024, 1, 2)]
    """
    day = start
    while day <= end:
        yield day
        day += datetime.timedelta(days=1)


def _iter_years(start, end):
    """Iter first days of years from start to end.

    The first item is the first day of the year of `start`.
    The last item is the first day of the year of `end`.

    >>> list(_iter_years(datetime.date(2022, 2, 20), datetime.date(2024, 11, 7)))
    [datetime.date(2022, 1, 1), datetime.date(2023, 1, 1), datetime.date(2024, 1, 1)]
    """
    year = start.replace(month=1, day=1)
    while year <= end:
        yield year
        year = (year + datetime.timedelta(days=366)).replace(month=1, day=1)


def get_history(period="day", /, *, dateformat=None):
    """Return the history of processed files.

    Returned value is a tuple of three lists `(dates, files, sizes)`, sorted by dates.
    For instance, `files[0]` files (totalizing `sizes[0]` bytes) have been processed on `dates[0]`.

    Dates may be formatted using `dateformat`.
    Otherwise, :class:`datetime.date` objects are returned.
    """
    if period == "year":
        history = {
            date.replace(month=1, day=1): (files, size)
            for (date, files, size) in db.session.query(
                func.min(DailyStat.date),
                func.sum(DailyStat.files),
                func.sum(DailyStat.size),
            ).group_by(strftime(db.engine.name, DailyStat.date, "%Y"))
        }
        # Fill in the blanks with (0, 0) on missing years
        history = [
            (date, *history.get(date, (0, 0)))
            for date in _iter_years(min(history.keys()), datetime.date.today())
        ]
    elif period == "month":
        history = {
            date.replace(day=1): (files, size)
            for (date, files, size) in db.session.query(
                func.min(DailyStat.date),
                func.sum(DailyStat.files),
                func.sum(DailyStat.size),
            ).group_by(strftime(db.engine.name, DailyStat.date, "%Y-%m"))
        }
        # Fill in the blanks with (0, 0) on missing months
        history = [
            (date, *history.get(date, (0, 0)))
            for date in _iter_months(min(history.keys()), datetime.date.today())
        ]
    elif period == "week":
        # History, where dates are stored as strings "YEAR-WEEK"
        strhistory = {
            date.strftime("%Y-%W"): (files, size)
            for (date, files, size) in db.session.query(
                func.min(DailyStat.date),
                func.sum(DailyStat.files),
                func.sum(DailyStat.size),
            ).group_by(strftime(db.engine.name, DailyStat.date, "%Y-%W"))
        }
        # Collate common weeks.
        # E.g. 2021-31-12 is week 52 of year 2021, 2022-01-01 is week 0 of year
        # 2022, but they are the very same week
        history = {}
        for strdate, data in strhistory.items():
            date = datetime.datetime.strptime(f"{strdate}-1", "%Y-%W-%w").date()
            if date in history:
                history[date] = (
                    history[date][0] + data[0],
                    history[date][1] + data[1],
                )
            else:
                history[date] = data
        # Fill in the blanks with (0, 0) on missing weeks
        history = [
            (date, *history.get(date, (0, 0)))
            for date in _iter_weeks(min(history.keys()), datetime.date.today())
        ]
    elif period == "day":
        history = {row.date: (row.files, row.size) for row in DailyStat.query.all()}
        # Fill in the blanks with (0, 0) on missing days
        history = [
            (date, *history.get(date, (0, 0)))
            for date in _iter_days(min(history.keys()), datetime.date.today())
        ]
    else:
        raise ValueError()

    if dateformat is None:
        return list(sorted(zip(*history)))
    return list(
        zip(
            *sorted(
                (date.strftime(dateformat), files, size)
                for date, files, size in history
            )
        )
    )



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Remove outdated files uploaded by users (original and imposed).

Run this from time to time (in a crontab for instance) if option ``RUN_CLEANER_THREAD`` is false.
"""

import sys

from flask import Flask

from .. import clean, settings


def main():
    """Main function: clean outdated files."""
    # Create a fake Flask application (to read settings)
    app = Flask("pdfimposeweb", instance_relative_config=True)
    settings.set_configuration(app)

    clean.clean(
        app.config["TEMP_FOLDER"],
        app.config["PDF_LIFETIME"],
    )


if __name__ == "__main__":
    if len(sys.argv) != 1:
        print(__doc__.strip())
        sys.exit(1)
    main()



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Generic functions and data about imposition"""

import importlib
import pathlib
import pkgutil
import sys

LAYOUTS = {
    name: getattr(module, "impose")
    for name, module in (
        (name, importlib.import_module(f".{name}", __name__))
        for finder, name, ispkg in pkgutil.iter_modules(__path__)
    )
    if hasattr(module, "impose")
}


def impose(layout, infile, outfile, arguments):
    """Dispatch arguments to the relevant imposition module."""
    LAYOUTS[layout](infile, outfile, arguments)



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Rate limiting."""

from flask_limiter import Limiter
from flask_limiter.util import get_remote_address


def init(app):
    """Initialize and return the limiter."""
    return Limiter(
        get_remote_address,
        app=app,
    )



# Copyright 2024 Louis Paternault

# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Hook to compile Babel catalog when building package."""

import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    """Hook that compile babel catalog when building a package."""

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        subprocess.run(
            "pybabel compile -d pdfimposeweb/translations", shell=True, check=True
        )



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Resize PDF

This is a test program. It may contain bugs and change without notice. Do not use it.

However, this could be easily be turned into a proper package.
If you think this might be useful, send me an email, it should not take me that long…
"""

import sys

from . import resize_scale, resize_size


def main():
    """Main function: parse arguments, and run script."""
    size, source, dest = sys.argv[1:]
    if size.startswith("x"):
        resize_scale(source, dest, float(size[1:]))
    elif size.startswith("/"):
        resize_scale(source, dest, 1 / float(size[1:]))
    else:
        resize_size(source, dest, size)


if __name__ == "__main__":
    main()



# Copyright 2024-2025 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Resize PDF (e.g. from A4 to A5)."""

import typing

import papersize
import pymupdf


def resize_scale(source: str, dest: str, scale: float):
    """Scale PDF (multiply or divide width and length by a scaling factor)"""
    sourcepdf = pymupdf.open(source)
    destpdf = pymupdf.open()

    for sourcepage in sourcepdf:
        x1, y1, x2, y2 = sourcepage.mediabox
        destpdf.new_page(
            width=scale * (x2 - x1), height=scale * (y2 - y1)
        ).show_pdf_page(
            pymupdf.Rect((0, 0), (scale * (x2 - x1), scale * (y2 - y1))),
            sourcepdf,
            sourcepage.number,
            keep_proportion=False,
            rotate=sourcepage.rotation,
        )

    destpdf.set_metadata(sourcepdf.metadata)  #  pylint: disable=no-member
    destpdf.save(dest)


def resize_size(source: str, dest: str, size: str | tuple[float]):
    """Resize PDF (to the size given in argument)"""
    sourcepdf = pymupdf.open(source)
    destpdf = pymupdf.open()

    if isinstance(size, str):
        size = map(float, papersize.parse_papersize(size))
    width, height = size

    for sourcepage in sourcepdf:
        destpdf.new_page(width=width, height=height).show_pdf_page(
            pymupdf.Rect((0, 0), (width, height)),
            sourcepdf,
            sourcepage.number,
            keep_proportion=False,
            rotate=sourcepage.rotation,
        )

    destpdf.set_metadata(sourcepdf.metadata)  #  pylint: disable=no-member
    destpdf.save(dest)



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Tools to process settings of the application."""

import pathlib

import flask


def set_configuration(app: flask.Flask):
    """Read configuration, and fix default options."""
    app.config.from_object("pdfimposeweb.settings.default")
    app.config.from_envvar("PDFIMPOSEWEB_SETTINGS", silent=True)

    # Fix default options
    if app.config["TEMP_FOLDER"] is None:
        app.config["TEMP_FOLDER"] = pathlib.Path(app.instance_path) / "temp"
    app.secret_key = app.config["SECRET_KEY"]



# Copyright 2024 Louis Paternault
#
# This file is part of pdfimpose-web.
#
# Pdfimpose-web is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Pdfimpose-web is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with pdfimpose-web. If not, see <https://www.gnu.org/licenses/>.

"""Perform "pdfautonup" imposition."""

import pdfautonup

from .common import parse_format, parse_length, parse_resize


def impose(infile, outfile, arguments):
    """Perform imposition.

    :param dict arguments: Arguments got from the POST html method.
    """
    parameters = {}

    # Paper size
    parameters["size"] = parse_format(arguments, "form-pdfautonup")[1]

    # Orientation
    parameters["orientation"] = arguments.get("form-pdfautonup-orientation", "auto")

    # Algorithm
    parameters["algorithm"] = arguments.get("form-pdfautonup-algo", "fuzzy")

    # Repeat
    if arguments.get("form-pdfautonup-repeat") in ("fit", "auto"):
        parameters["repeat"] = arguments.get("form-pdfautonup-repeat")
    else:
        try:
            parameters["repeat"] = int(arguments.get("form-pdfautonup-repeat-value"))
        except ValueError:
            parameters["repeat"] = 1

    # Margin and gap
    if parameters["algorithm"] == "panel":
        parameters["more"] = {
            "margin": parse_length(arguments, "form-pdfautonup-panel", "margin"),
            "gap": parse_length(arguments, "form-pdfautonup-panel", "gap"),
        }

    # Argument "resize"
    infile = parse_resize(infile, arguments, "form-pdfautonup")

    pdfautonup.pdfautonup(
        files=[infile],
        output=outfile,
        **parameters,
    )
